from ai2_kit.core.artifact import Artifact
from ai2_kit.core.script import BashTemplate
from ai2_kit.core.executor import Executor
from ai2_kit.core.script import BashScript, BashStep
from ai2_kit.core.job import JobFuture, GatherJobsFuture, retry_fn, map_future
from ai2_kit.core.log import get_logger

from pydantic import BaseModel
from typing import List
from dataclasses import dataclass
import os
import copy
import random
import sys
import json

from .constant import (
    DP_CHECKPOINT_FILE,
    DP_DISP_FILE,
    DP_PROFILING_FILE,
    DP_INPUT_FILE,
    DP_FROZEN_MODEL,
)

logger = get_logger(__name__)


class GeneralDeepmdInputConfig(BaseModel):
    model_num: int = 4
    init_data: List[str]
    input_template: dict


class GeneralDeepmdContextConfig(BaseModel):
    script_template: BashTemplate
    dp_cmd: str = 'dp'


@dataclass
class GeneralDeepmdInput:
    config: GeneralDeepmdInputConfig
    type_map: List[str]
    old_data: List[Artifact]  # training data used by previous iteration
    new_data: List[Artifact]  # training data used by current iteration

@dataclass
class GeneralDeepmdContext:
    path_prefix: str
    executor: Executor
    config: GeneralDeepmdContextConfig

@dataclass
class GeneralDeepmdOutput:
    @dataclass
    class Result:
        output_dir: Artifact

    results: List[Result]
    set_prefix: str



def general_deepmd(input: GeneralDeepmdInput, ctx: GeneralDeepmdContext):

    jobs: List[JobFuture] = []
    results: List[GeneralDeepmdOutput.Result] = []

    # train multiple models at the same time
    for i in range(input.config.model_num):
        # each model should be trained in its own work_dir
        path_prefix = os.path.join(ctx.path_prefix, str(i).zfill(4))
        work_dir = ctx.executor.get_full_path(path_prefix)
        ctx.executor.mkdir(work_dir)
        logger.info('the work_dir is %s', work_dir)

        # create dp train input file
        # NOTE: dp v1 format is supported currently
        # TODO: migrate to dp v2 format
        # TODO: support more params if it is necessary
        # TODO: support train from previous model

        # ref: https://github.com/deepmodeling/dpgen2/blob/master/examples/ch4/param_CH4_deepmd-kit-2.1.1.json
        # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/prep_dp_train.py
        # ref: https://github.com/deepmodeling/dpgen2/blob/master/dpgen2/op/run_dp_train.py

        dp_input = copy.deepcopy(input.config.input_template)
        training: dict = dp_input['training']

        # set output files
        training['disp_file'] = DP_DISP_FILE
        training['save_ckpt'] = DP_CHECKPOINT_FILE
        training['profiling_file'] = DP_PROFILING_FILE

        # set random seed
        discriptor = dp_input['model']['descriptor']
        if discriptor['type'] == 'hybrid':
            for d in discriptor['list']:
                d['seed'] = _random_seed()
        else:
            discriptor['seed'] = _random_seed()
        dp_input['model']['fitting_net']['seed'] = _random_seed()
        dp_input['training']['seed'] = _random_seed()

        # set training data
        systems = [a.url for a in input.old_data + input.new_data]
        training['systems'] = systems

        # set other params
        dp_input['model']['type_map'] = input.type_map
        set_prefix: str = training.setdefault(
            'set_prefix', 'set')  # respect user input
        auto_prob_str = "prob_sys_size"
        training.setdefault('batch_size', 'auto')
        training['auto_prob_style'] = auto_prob_str

        # write config to executor
        dp_input_text = json.dumps(dp_input, indent=2)
        dp_input_path = os.path.join(work_dir, DP_INPUT_FILE)
        ctx.executor.dump_text(dp_input_text, dp_input_path)

        # build script
        dp_cmd = ctx.config.dp_cmd
        dp_train_cmd = [dp_cmd, 'train', DP_INPUT_FILE]
        dp_freeze_cmd = [dp_cmd, 'freeze', '-o', DP_FROZEN_MODEL]

        dp_train_script = BashScript(
            template=ctx.config.script_template,
            steps=[
                BashStep(cmd=dp_train_cmd, checkpoint='dp-train'),  # type: ignore
                BashStep(cmd=dp_freeze_cmd),  # type: ignore
            ] # type: ignore
        )

        results.append(GeneralDeepmdOutput.Result(
            output_dir=Artifact(
                executor=ctx.executor.name,
                url=work_dir,
                attrs=dict(
                    frozen_model=DP_FROZEN_MODEL,
                )
            ) # type: ignore
        ))

        # submit job
        job = ctx.executor.submit(dp_train_script.render(), cwd=work_dir)

        # FIXME: investigate this type issue
        jobs.append(job)  # type: ignore

    future = GatherJobsFuture(jobs, done_fn=retry_fn(max_tries=2), raise_exception=True)

    return map_future(future, GeneralDeepmdOutput(
        set_prefix=set_prefix,  # type: ignore
        results=results,
    ))

def _random_seed():
    return random.randrange(sys.maxsize) % (1 << 32)



from ai2_kit.core.script import BashTemplate, BashStep, BashScript
from ai2_kit.core.artifact import Artifact
from ai2_kit.core.executor import Executor
from ai2_kit.core.log import get_logger
from ai2_kit.core.job import GatherJobsFuture, retry_fn, map_future

from typing import List, Literal, Optional
from pydantic import BaseModel
from dataclasses import dataclass
from string import Template
import os
import math
import itertools
import random

from .constant import (
    MODEL_DEVI_OUT,
    MODEL_DEVI_NEU_OUT,
    MODEL_DEVI_RED_OUT,
    LAMMPS_TRAJ_DIR,
    LAMMPS_TRAJ_SUFFIX,
)

logger = get_logger(__name__)

class MdVariants(BaseModel):
    system_vars: List[str]  # vars = variants
    temp_vars: List[int]
    pres_vars: List[float]
    lambda_vars: Optional[List[float]]
    trj_freq: int
    nsteps: int
    ensemble: Literal['nvt', 'nvt-i', 'nvt-a', 'nvt-iso', 'nvt-aniso', 'npt', 'npt-t', 'npt-tri', 'nve']


class GeneralLammpsInputConfig(BaseModel):
    iters: List[MdVariants]  # Explore Options for each iteration
    no_pbc: bool = False
    tau_t: float = 0.1
    tau_p: float = 0.5
    timestep: float = 0.0005
    input_template: Optional[str]


class GeneralLammpsContextConfig(BaseModel):
    script_template: BashTemplate
    lammps_cmd: str = 'lmp'
    concurrency: int = 5


@dataclass
class GeneralLammpsInput:

    @dataclass
    class MdOptions:
        models: List[Artifact]

    @dataclass
    class FepOptions:
        red_models: List[Artifact]
        neu_models: List[Artifact]

    config: GeneralLammpsInputConfig
    type_map: List[str]
    mass_map: List[float]

    md_vars: MdVariants
    system_vars: List[Artifact]

    # The following options are mutex
    md_options: Optional[MdOptions] = None
    fep_options: Optional[FepOptions] = None


@dataclass
class GeneralLammpsContext:
    config: GeneralLammpsContextConfig
    path_prefix: str
    executor: Executor

@dataclass
class GeneralLammpsOutput:
    candidates: List[Artifact]

def general_lammps(input: GeneralLammpsInput, ctx: GeneralLammpsContext):
    """
    1. resolve system file path
    2. convert system into lammps format
    3. build configuration file
    4. build script and submit
    """

    # 0. setup dirs
    work_dir = ctx.executor.get_full_path(ctx.path_prefix)
    ctx.executor.mkdir(work_dir)
    logger.info('work_dir is %s', work_dir)

    input_data_dir = os.path.join(work_dir, 'input_data')
    ctx.executor.mkdir(input_data_dir)
    logger.info('data_dir is %s', input_data_dir)

    tasks_dir = os.path.join(work_dir, 'tasks')
    ctx.executor.mkdir(tasks_dir)
    logger.info('tasks_dir is %s', tasks_dir)

    # 1. resolve file path
    system_files_set = set()
    for system_opt in input.system_vars:
        system_files_set.update(ctx.executor.unpack_artifact(system_opt))
    system_files = sorted(system_files_set)

    # 2. convert system into lammps format
    lammps_data_files = []
    zfill_size = math.ceil(math.log10(len(system_files)))
    for i, sys_file in enumerate(system_files):
        file_name = f'{str(i).zfill(zfill_size)}-{os.path.basename(sys_file)}.lammps.data'
        lammps_data_files.append(os.path.join(input_data_dir, file_name))

    def to_lammps_input(input_files: List[str], output_files: List[str], type_map: List[str], fmt='vasp/poscar'):
        import dpdata
        for i in range(len(input_files)):
            dpdata.System(input_files[i], fmt=fmt, type_map=type_map).to_lammps_lmp(output_files[i])  # type: ignore

    ctx.executor.run_python_fn(to_lammps_input)(system_files, lammps_data_files, input.type_map)

    # 3. build configuration files
    md_vars = input.md_vars

    lammps_task_dirs = []
    lammps_input_file_name = 'lammps.input'
    task_counter = itertools.count(0)
    for lammps_data_file in lammps_data_files:
        for temp in md_vars.temp_vars:
            for pres in md_vars.pres_vars:
                task_num = next(task_counter)
                lammps_task_dir = os.path.join(tasks_dir, str(task_num).zfill(zfill_size + 3))  # TODO: optimize zfill
                ctx.executor.mkdir(os.path.join(lammps_task_dir, LAMMPS_TRAJ_DIR))  # create dump directory for lammps or else will get error
                lambda_vars = md_vars.lambda_vars
                lambda_f = lambda_vars[task_num % len(lambda_vars)] if lambda_vars else None

                # TODO: ensure only one of md, fep is set
                if input.md_options:
                    models = [a.url for a in input.md_options.models]
                    force_field_settings = make_md_force_field_settings(models=models)
                    template = input.config.input_template or DEFAULT_MD_INPUT_TEMPLATE

                elif input.fep_options:
                    if lambda_f is None:
                        raise ValueError('lambda_vars must be set when fep is used!')
                    neu_models = [a.url for a in input.fep_options.neu_models]
                    red_models = [a.url for a in input.fep_options.red_models]
                    force_field_settings = make_fep_force_field_settings(neu_models=neu_models, red_models=red_models)
                    template = input.config.input_template or DEFAULT_FEP_INPUT_TEMPLATE
                else:
                    raise ValueError('one and only one of md_options or fep_options must be set')

                input_text = make_lammps_input(
                    template=template,
                    data_file=lammps_data_file,
                    temp=temp,
                    pres=pres,
                    lambda_f=lambda_f,
                    force_field_settings=force_field_settings,

                    nsteps=md_vars.nsteps,
                    trj_freq=md_vars.trj_freq,
                    ensemble=md_vars.ensemble,

                    tau_p=input.config.tau_p,
                    tau_t=input.config.tau_t,
                    timestep=input.config.timestep,

                    mass_map=input.mass_map,
                )
                input_file_path = os.path.join(lammps_task_dir, lammps_input_file_name)
                ctx.executor.dump_text(input_text, input_file_path)
                lammps_task_dirs.append(lammps_task_dir)

    # 4. build scripts and submit
    lammps_cmd = ctx.config.lammps_cmd
    base_cmd = f'{lammps_cmd} -i {lammps_input_file_name}'
    cmd = f'''if [ -f md.restart.10000 ]; then {base_cmd} -v restart 1; else {base_cmd} -v restart 0; fi'''

    # group tasks by concurrency
    concurrency = ctx.config.concurrency
    steps_group = [ list() for _ in range(concurrency)]

    for i, lammps_task_dir in enumerate(lammps_task_dirs):
        steps = steps_group[i % concurrency]
        step = BashStep(
            cwd=lammps_task_dir,
            cmd=cmd,
            checkpoint='lammps',
        )
        steps.append(step)

    # submit jobs
    jobs = []
    for steps in steps_group:
        if not steps:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps,
        )
        job = ctx.executor.submit(script.render(), cwd=tasks_dir)
        jobs.append(job)

    outputs = [
        Artifact(
            url=task_dir,
            executor=ctx.executor.name,
            attrs=dict(by='lammps'),
        ) for task_dir in lammps_task_dirs]  # type: ignore

    future = GatherJobsFuture(jobs, done_fn=retry_fn(max_tries=2), raise_exception=True)
    return map_future(future, GeneralLammpsOutput(
        candidates=outputs
    ))


def make_md_force_field_settings(models: List[str]):
    deepmd_args = ""
    settings = [
        'pair_style deepmd %s out_freq ${THERMO_FREQ} out_file %s %s' % (' '.join(models), MODEL_DEVI_OUT, deepmd_args),
        'pair_coeff * *',
    ]
    return settings


def make_fep_force_field_settings(neu_models: List[str], red_models: List[str]):
    deepmd_args = ""
    settings = [
        'pair_style hybrid/overlay &',
        '           deepmd %s out_freq ${THERMO_FREQ} out_file %s %s &' %(' '.join(neu_models), MODEL_DEVI_NEU_OUT, deepmd_args),
        '           deepmd %s out_freq ${THERMO_FREQ} out_file %s %s'   %(' '.join(red_models), MODEL_DEVI_RED_OUT, deepmd_args),
        'pair_coeff  * * deepmd 1 *',
        'pair_coeff  * * deepmd 2 *',
        '',
        'fix sampling_PES all adapt 0 &',
        '    pair deepmd:1 scale * * v_LAMBDA_f &',
        '    pair deepmd:2 scale * * v_LAMBDA_i &',
        '    scale yes',
    ]
    return settings


def make_lammps_input(template: str,
                      data_file: str,
                      nsteps: int,
                      timestep: float,
                      trj_freq: int,
                      temp: float,
                      pres: float,
                      tau_t: float,
                      tau_p: float,
                      ensemble: str,
                      mass_map: List[float],
                      force_field_settings: List[str],
                      lambda_f: Optional[float] = None,
                      no_pbc=False,
                      rand_start=1_000_000,
                      ):

    # FIXME: Is it a good idea to fix pres or just raise error?
    if not ensemble.startswith('npt'):
        pres = -1

    variables = [
        'variable NSTEPS      equal %d' % nsteps,
        'variable THERMO_FREQ equal %d' % trj_freq,
        'variable DUMP_FREQ   equal %d' % trj_freq,
        'variable TEMP        equal %f' % temp,
        'variable PRES        equal %f' % pres,
        'variable TAU_T       equal %f' % tau_t,
        'variable TAU_P       equal %f' % tau_p,
    ]

    if lambda_f is not None:
        variables.extend([
            '',
            'variable LAMBDA_f    equal %.2f' % lambda_f,
            'variable LAMBDA_i    equal 1-v_LAMBDA_f',
        ])

    init_settings = [
        'boundary ' + ('f f f' if no_pbc else 'p p p'),
    ]

    atom_settings = [
        '''if "${restart} > 0" then "read_restart md.restart.*" else "read_data %s"''' % data_file,
        *("mass {id} {mass}".format(id=i+1, mass=m) for i, m in enumerate(mass_map))
    ]

    md_settings = [
        '''if "${restart} == 0" then "velocity all create ${TEMP} %d"''' % (random.randrange(rand_start - 1) + 1)
    ]

    if ensemble.startswith('npt') and no_pbc:
        raise ValueError('ensemble npt conflict with no_pcb')
    if ensemble in ('npt', 'npt-i', 'npt-iso',):
        md_settings.append('fix 1 all npt temp ${TEMP} ${TEMP} ${TAU_T} iso ${PRES} ${PRES} ${TAU_P}')
    elif ensemble in ('npt-a', 'npt-aniso',):
        md_settings.append('fix 1 all npt temp ${TEMP} ${TEMP} ${TAU_T} aniso ${PRES} ${PRES} ${TAU_P}')
    elif ensemble in ('npt-t', 'npt-tri',):
        md_settings.append('fix 1 all npt temp ${TEMP} ${TEMP} ${TAU_T} tri ${PRES} ${PRES} ${TAU_P}')
    elif ensemble in ('nvt',):
        md_settings.append('fix 1 all nvt temp ${TEMP} ${TEMP} ${TAU_T}')
    elif ensemble in ('nve',):
        md_settings.append('fix 1 all nve')
    else:
        raise ValueError('unknown ensemble: ' + ensemble)

    if no_pbc:
        md_settings.extend([
            'velocity all zero linear',
            'fix      fm all momentum 1 linear 1 1 1',
        ])

    md_settings.extend([
        'thermo_style custom step temp pe ke etotal press vol lx ly lz xy xz yz',
        'thermo       ${THERMO_FREQ}',
        'dump         1 all custom ${DUMP_FREQ} %s/*%s id type x y z fx fy fz' % (LAMMPS_TRAJ_DIR, LAMMPS_TRAJ_SUFFIX),
        'restart      10000 md.restart',
    ])

    run_settings = [
        'timestep %f' % timestep,
        'run      ${NSTEPS} upto',
    ]

    return LammpsInputTemplate(template).substitute(dict(
        variable_settings='\n'.join(variables),
        init_settings='\n'.join(init_settings),
        atom_settings='\n'.join(atom_settings),
        md_settings='\n'.join(md_settings),
        force_field_settings='\n'.join(force_field_settings),
        run_settings='\n'.join(run_settings),
    ))


class LammpsInputTemplate(Template):
    """
    change delimiter from $ to $$ as $ is used a lot in lammps input file
    """
    delimiter = '$$'


DEFAULT_MD_INPUT_TEMPLATE = '''# MD INPUT

## VARIABLES
$$variable_settings

## INITIALIZATION SETTINGS
units      metal
atom_style atomic
neighbor   1.0 bin
box        tilt large

$$init_settings

## ATOM SETTINGS
$$atom_settings

change_box all triclinic

## FORCE FIELD SETTINGS
$$force_field_settings

## MD SETTINGS
$$md_settings

## RUN SETTINGS
$$run_settings
'''

DEFAULT_FEP_INPUT_TEMPLATE = '''# FEP INPUT

## VARIABLES
$$variable_settings

variable plus  equal 1
variable minus equal -1

## INITIALIZATION SETTINGS
units        metal
atom_style   atomic
atom_modify  map yes
neighbor     2.0 bin
neigh_modify every 1

$$init_settings

## ATOM SETTINGS
$$atom_settings

change_box all triclinic

## FORCE FIELD SETTINGS
$$force_field_settings

## MD SETTINGS
$$md_settings

## RUN SETTINGS
$$run_settings
'''


from ai2_kit.core.executor import BaseExecutorConfig, ExecutorManager
from ai2_kit.core.artifact import ArtifactMap
from ai2_kit.core.log import get_logger
from ai2_kit.core.util import load_yaml_files
from ai2_kit.domain import deepmd
from ai2_kit.domain import lammps
from ai2_kit.domain import selector
from ai2_kit.domain import cp2k
from ai2_kit.domain import constant as const

from pydantic import BaseModel
from typing import Dict, List, Optional
from fire import Fire

import os

logger = get_logger(__name__)


class CllWorkflowExecutorConfig(BaseExecutorConfig):
    class Context(BaseModel):
        class Train(BaseModel):
            deepmd: deepmd.GeneralDeepmdContextConfig

        class Explore(BaseModel):
            lammps: lammps.GeneralLammpsContextConfig

        class Label(BaseModel):
            cp2k: cp2k.GeneralCp2kContextConfig

        train: Train
        explore: Explore
        label: Label

    context: Context


class CllWorkflowConfig(BaseModel):
    class Workflow(BaseModel):
        class General(BaseModel):
            type_map: List[str]
            mass_map: List[float]
            max_iters: int = 10

        class Train(BaseModel):
            deepmd: deepmd.GeneralDeepmdInputConfig

        class Explore(BaseModel):
            lammps: lammps.GeneralLammpsInputConfig

        class Select(BaseModel):
            threshold: selector.GeneralThresholdInputConfig

        class Label(BaseModel):
            cp2k: cp2k.GeneralCp2kInputConfig

        general: General
        train: Train
        explore: Explore
        select: Select
        label: Label

    executors: Dict[str, CllWorkflowExecutorConfig]
    artifacts: ArtifactMap
    workflow: Workflow


def cll_train_mlp(*config_files, executor: Optional[str] = None, path_prefix: Optional[str] = None):
    """
    Run Closed-Loop Learning (CLL) workflow to train Machine Learning Potential (MLP) models.
    """

    config_data = load_yaml_files(*config_files)
    config = CllWorkflowConfig.parse_obj(config_data)

    if executor not in config.executors:
        raise ValueError(f'executor {executor} is not found')
    if path_prefix is None:
        raise ValueError('path_prefix should not be empty')

    executor_manager = ExecutorManager(config.executors)  # type: ignore
    default_executor = executor_manager.get_executor(executor)

    context_cfg = config.executors[executor].context

    type_map = config.workflow.general.type_map
    mass_map = config.workflow.general.mass_map

    # Init Setting
    deepmd_input = deepmd.GeneralDeepmdInput(
        config=config.workflow.train.deepmd,
        type_map=type_map,
        old_data=[],
        new_data=[config.artifacts[key] for key in config.workflow.train.deepmd.init_data],
    )
    deepmd_context = deepmd.GeneralDeepmdContext(
        path_prefix='',
        executor=default_executor,
        config=context_cfg.train.deepmd,
    )

    # Start iteration
    for i in range(config.workflow.general.max_iters):
        # update path prefix for each iteration
        iter_path_prefix = os.path.join(path_prefix, f'iters-{str(i).zfill(3)}')

        # TODO: change order to: label -> train -> explore -> select
        # TODO: support more tools
        # TODO: support more options

        # train
        deepmd_context.path_prefix = os.path.join(iter_path_prefix, 'train-deepmd')
        deepmd_output = deepmd.general_deepmd(deepmd_input, deepmd_context).result()

        # explore
        models = [a.output_dir.join(a.output_dir.attrs['frozen_model']) for a in deepmd_output.results]
        lammps_iters = config.workflow.explore.lammps.iters
        md_vars = lammps_iters[ i if i < len(lammps_iters) else (len(lammps_iters) - 1) ]

        lammps_input = lammps.GeneralLammpsInput(
            config=config.workflow.explore.lammps,
            type_map=type_map,
            mass_map=mass_map,
            md_vars=md_vars,
            system_vars=[config.artifacts[key] for key in md_vars.system_vars],
            md_options=lammps.GeneralLammpsInput.MdOptions(models=models)
        )
        lammps_context = lammps.GeneralLammpsContext(
            config=context_cfg.explore.lammps,
            path_prefix=os.path.join(iter_path_prefix, 'explore-lammps'),
            executor=default_executor,
        )
        lammps_output = lammps.general_lammps(lammps_input, lammps_context).result()

        # select
        selector_input = selector.LammpsGeneralThresholdInput(
            config=config.workflow.select.threshold,
            candidates=lammps_output.candidates,
            model_devi_out_file=const.MODEL_DEVI_OUT,
        )
        selector_context = selector.GeneralThresholdContext(
            executor=default_executor,
        )
        selector_output = selector.lammps_general_threshold(selector_input, selector_context).result()

        # label
        cp2k_input = cp2k.GeneralCp2kInput(
            config=config.workflow.label.cp2k,
            type_map=config.workflow.general.type_map,
            candidates=selector_output.candidates,
            basic_set_file=config.artifacts.get(config.workflow.label.cp2k.basic_set_file or ''),
            potential_file=config.artifacts.get(config.workflow.label.cp2k.potential_file or ''),
        )
        cp2k_context = cp2k.GeneralCp2kContext(
            config=context_cfg.label.cp2k,
            path_prefix=os.path.join(iter_path_prefix, 'label-cp2k'),
            executor=default_executor,
        )
        cp2k_output = cp2k.general_cp2k(cp2k_input, cp2k_context).result()

        # update input for the next round
        deepmd_input.old_data.extend(deepmd_input.new_data)
        deepmd_input.new_data = cp2k_output.dp_data_sets

        # deepmd_input, lammps_input, selector_input, cp2k_input = smart_update(
        #     i,
        #     deepmd_output, lammps_output, selector_output, cp2k_output,
        #     deepmd_input, lammps_input, selector_input, cp2k_input,
        # )



if __name__ == '__main__':
    # for test, e.g:
    Fire(cll_train_mlp)



from ruamel.yaml import YAML
from pathlib import Path
from typing import Tuple
import shortuuid
import copy
from dataclasses import field

def default_mutable_field(obj):
    return field(default_factory=lambda: copy.copy(obj))

def __merge_dict():
    """
    merge two dict, the left dict will be overridden
    this method won't merge list

    cloudpickle compatible: https://stackoverflow.com/questions/75292769
    """
    def merge_dict(lo: dict, ro: dict, path=None):
        if path is None:
            path = []

        for key, value in ro.items():
            if key in lo:
                current_path = path + [ str(key) ]
                if isinstance(lo[key], dict) and isinstance(value, dict):
                    merge_dict(lo[key], value, current_path)
                else:
                    print('.'.join(current_path) + ' has been overridden')
                    lo[key] = value
            else:
                lo[key] = value
        return lo
    return merge_dict
merge_dict = __merge_dict()


# TODO: support http(s) url
def load_yaml_file(path: Path):
    yaml = YAML(typ='safe')
    return yaml.load(path)


def load_yaml_files(*paths: Tuple[Path]):
    d = {}
    for path in paths:
        print('load yaml file: ', path)
        d = merge_dict(d, load_yaml_file(Path(path)))
    return d


def tmpfile_name():
    return shortuuid.uuid()



from ai2_kit.core.executor import BaseExecutorConfig, ExecutorManager
from ai2_kit.core.artifact import ArtifactMap
from ai2_kit.core.log import get_logger
from ai2_kit.core.util import load_yaml_files
from ai2_kit.domain import deepmd
from ai2_kit.domain import lammps
from ai2_kit.domain import selector
from ai2_kit.domain import cp2k
from ai2_kit.domain import constant as const

from pydantic import BaseModel
from typing import Dict, List, Optional
from fire import Fire

import os

logger = get_logger(__name__)


class FepExecutorConfig(BaseExecutorConfig):
    class Context(BaseModel):

        deepmd: deepmd.GeneralDeepmdContextConfig
        lammps: lammps.GeneralLammpsContextConfig
        cp2k: cp2k.GeneralCp2kContextConfig

    context: Context


class FepConfig(BaseModel):
    class Workflow(BaseModel):
        class General(BaseModel):
            type_map: List[str]
            mass_map: List[float]
            max_iters: int = 10

        class Branch(BaseModel):
            deepmd: deepmd.GeneralDeepmdInputConfig
            cp2k: cp2k.GeneralCp2kInputConfig
            threshold: selector.GeneralThresholdInputConfig

        general: General
        neu: Branch
        red: Branch
        lammps: lammps.GeneralLammpsInputConfig

    executors: Dict[str, FepExecutorConfig]
    artifacts: ArtifactMap
    workflow: Workflow

def fep_train_mlp(*config_files, executor: Optional[str] = None, path_prefix: Optional[str] = None):
    """
    Training ML potential for FEP
    """

    config_data = load_yaml_files(*config_files)
    config = FepConfig.parse_obj(config_data)

    if executor not in config.executors:
        raise ValueError(f'executor {executor} is not found')
    if path_prefix is None:
        raise ValueError('path_prefix should not be empty')

    executor_manager = ExecutorManager(config.executors)  # type: ignore
    default_executor = executor_manager.get_executor(executor)

    context_cfg = config.executors[executor].context

    type_map = config.workflow.general.type_map
    mass_map = config.workflow.general.mass_map

    # Init Setting
    neu_deepmd_input = deepmd.GeneralDeepmdInput(
        config=config.workflow.neu.deepmd,
        type_map=type_map,
        old_data=[],
        new_data=[config.artifacts[key] for key in config.workflow.neu.deepmd.init_data],
    )
    neu_deepmd_context = deepmd.GeneralDeepmdContext(
        path_prefix='',
        executor=default_executor,
        config=context_cfg.deepmd,
    )

    red_deepmd_input = deepmd.GeneralDeepmdInput(
        config=config.workflow.red.deepmd,
        type_map=type_map,
        old_data=[],
        new_data=[config.artifacts[key] for key in config.workflow.red.deepmd.init_data],
    )
    red_deepmd_context = deepmd.GeneralDeepmdContext(
        path_prefix='',
        executor=default_executor,
        config=context_cfg.deepmd,
    )

    # Start iteration
    for i in range(config.workflow.general.max_iters):
        # update path prefix for each iteration
        iter_path_prefix = os.path.join(path_prefix, f'iters-{str(i).zfill(3)}')

        # train
        neu_deepmd_context.path_prefix = os.path.join(iter_path_prefix, 'train-neu-deepmd')
        neu_deepmd_output_future = deepmd.general_deepmd(neu_deepmd_input, neu_deepmd_context)

        red_deepmd_context.path_prefix = os.path.join(iter_path_prefix, 'train-red-deepmd')
        red_deepmd_output_future = deepmd.general_deepmd(red_deepmd_input, red_deepmd_context)

        neu_deepmd_output, red_deepmd_output = neu_deepmd_output_future.result(), red_deepmd_output_future.result()

        # explore
        neu_models = [a.output_dir.join(a.output_dir.attrs['frozen_model']) for a in neu_deepmd_output.results]
        red_models = [a.output_dir.join(a.output_dir.attrs['frozen_model']) for a in red_deepmd_output.results]

        lammps_iters = config.workflow.lammps.iters
        md_vars = lammps_iters[i % len(lammps_iters)]

        lammps_input = lammps.GeneralLammpsInput(
            config=config.workflow.lammps,
            type_map=type_map,
            mass_map=mass_map,
            md_vars=md_vars,
            system_vars=[config.artifacts[key] for key in md_vars.system_vars],
            fep_options=lammps.GeneralLammpsInput.FepOptions(red_models=red_models, neu_models=neu_models)
        )
        lammps_context = lammps.GeneralLammpsContext(
            config=context_cfg.lammps,
            path_prefix=os.path.join(iter_path_prefix, 'explore-lammps'),
            executor=default_executor,
        )
        lammps_output = lammps.general_lammps(lammps_input, lammps_context).result()

        # select
        neu_selector_input = selector.LammpsGeneralThresholdInput(
            config=config.workflow.neu.threshold,
            candidates=lammps_output.candidates,
            model_devi_out_file=const.MODEL_DEVI_NEU_OUT,
        )
        neu_selector_context = selector.GeneralThresholdContext(
            executor=default_executor,
        )

        red_selector_input = selector.LammpsGeneralThresholdInput(
            config=config.workflow.red.threshold,
            candidates=lammps_output.candidates,
            model_devi_out_file=const.MODEL_DEVI_RED_OUT,
        )
        red_selector_context = selector.GeneralThresholdContext(
            executor=default_executor,
        )

        neu_selector_output_future = selector.lammps_general_threshold(neu_selector_input, neu_selector_context)
        red_selector_output_future = selector.lammps_general_threshold(red_selector_input, red_selector_context)

        neu_selector_output, red_selector_output = neu_selector_output_future.result(), red_selector_output_future.result()

        # label
        neu_cp2k_input = cp2k.GeneralCp2kInput(
            config=config.workflow.neu.cp2k,
            type_map=config.workflow.general.type_map,
            candidates=neu_selector_output.candidates,
            basic_set_file=config.artifacts.get(config.workflow.neu.cp2k.basic_set_file or ''),
            potential_file=config.artifacts.get(config.workflow.neu.cp2k.potential_file or ''),
        )
        neu_cp2k_context = cp2k.GeneralCp2kContext(
            config=context_cfg.cp2k,
            path_prefix=os.path.join(iter_path_prefix, 'label-neu-cp2k'),
            executor=default_executor,
        )

        red_cp2k_input = cp2k.GeneralCp2kInput(
            config=config.workflow.red.cp2k,
            type_map=config.workflow.general.type_map,
            candidates=red_selector_output.candidates,
            basic_set_file=config.artifacts.get(config.workflow.red.cp2k.basic_set_file or ''),
            potential_file=config.artifacts.get(config.workflow.red.cp2k.potential_file or ''),
        )
        red_cp2k_context = cp2k.GeneralCp2kContext(
            config=context_cfg.cp2k,
            path_prefix=os.path.join(iter_path_prefix, 'label-red-cp2k'),
            executor=default_executor,
        )

        neu_cp2k_output_future = cp2k.general_cp2k(neu_cp2k_input, neu_cp2k_context)
        red_cp2k_output_future = cp2k.general_cp2k(red_cp2k_input, red_cp2k_context)

        neu_cp2k_output, red_cp2k_output = neu_cp2k_output_future.result(), red_cp2k_output_future.result()

        # update input for the next round
        neu_deepmd_input.old_data.extend(neu_deepmd_input.new_data)
        neu_deepmd_input.new_data = neu_cp2k_output.dp_data_sets

        red_deepmd_input.old_data.extend(red_deepmd_input.new_data)
        red_deepmd_input.new_data = red_cp2k_output.dp_data_sets


if __name__ == '__main__':
    # for test, e.g:
    # python -m ai2_kit.workflow.fep fep_config.yaml --executor=hpc01 --path-prefix=fep-test
    Fire(fep_train_mlp)


from ai2_kit.core.executor import Executor
from ai2_kit.core.artifact import Artifact
from ai2_kit.core.script import BashScript, BashStep, BashTemplate
from ai2_kit.core.job import GatherJobsFuture, MapFuture, retry_fn, DummyFuture, IFuture

from ai2_kit.core.util import merge_dict
from ai2_kit.core.log import get_logger


from typing import List, Optional
from pydantic import BaseModel
from dataclasses import dataclass

import copy
import os

from .constant import (
    LAMMPS_TRAJ_DIR,
    LAMMPS_TRAJ_SUFFIX,
)

logger = get_logger(__name__)

class GeneralCp2kInputConfig(BaseModel):
    max_candidates: int = 50
    input_template: dict
    basic_set_file: Optional[str]
    potential_file: Optional[str]

class GeneralCp2kContextConfig(BaseModel):
    script_template: BashTemplate
    cp2k_cmd: str = 'cp2k'
    concurrency: int = 5

@dataclass
class GeneralCp2kInput:
    config: GeneralCp2kInputConfig
    candidates: List[Artifact]
    type_map: List[str]

    basic_set_file: Optional[Artifact] = None
    potential_file: Optional[Artifact] = None

@dataclass
class GeneralCp2kContext:
    config: GeneralCp2kContextConfig
    executor: Executor
    path_prefix: str

@dataclass
class GeneralCp2kOutput:
    dp_data_sets: List[Artifact]
    cp2k_outputs: List[Artifact]

def general_cp2k(input: GeneralCp2kInput, ctx: GeneralCp2kContext) -> IFuture[GeneralCp2kOutput]:
    # setup dirs
    work_dir = ctx.executor.get_full_path(ctx.path_prefix)
    ctx.executor.mkdir(work_dir)
    logger.info('work_dir is %s', work_dir)

    tasks_dir = os.path.join(work_dir, 'tasks')
    ctx.executor.mkdir(tasks_dir)
    logger.info('tasks_dir is %s', tasks_dir)

    dp_data_dir = os.path.join(work_dir, 'output_dp_data')
    ctx.executor.mkdir(dp_data_dir)
    logger.info('dp_data_dir is %s', dp_data_dir)

    # resolve input template
    input_template = copy.deepcopy(input.config.input_template)
    # FIXME: potential_file and basic_set_file should be list
    if input.basic_set_file is not None:
        merge_dict(input_template,  {
            'FORCE_EVAL': {
                'DFT': {
                    'BASIS_SET_FILE_NAME': input.basic_set_file.url,
                }
            }
        })
    if input.potential_file is not None:
        merge_dict(input_template,  {
            'FORCE_EVAL': {
                'DFT': {
                    'POTENTIAL_FILE_NAME': input.potential_file.url,
                }
            }
        })

    # resolve candidate files
    lammps_traj_files = []

    for candidate in input.candidates:
        by = candidate.attrs['by']
        if by == 'lammps':
            traj_files = [os.path.join(candidate.url, LAMMPS_TRAJ_DIR, f'{i}{LAMMPS_TRAJ_SUFFIX}') for i in candidate.attrs['passed']]
            lammps_traj_files.extend(traj_files)
        else:
            logger.warn('skip unsupported data source: %s, generated by: %s', candidate.url, by)
            continue

    # limit max_candidates
    # TODO: allow user to choose random sample
    lammps_traj_files = lammps_traj_files[:input.config.max_candidates]

    # create task dirs and prepare input files
    cp2k_task_dirs = []
    if lammps_traj_files:
        cp2k_task_dirs = [
            os.path.join(tasks_dir, str(i).zfill(6))
            for i in range(len(lammps_traj_files))
        ]
        ctx.executor.run_python_fn(make_cp2k_task_dirs)(
            task_dirs=cp2k_task_dirs,
            traj_files=lammps_traj_files,
            traj_fmt='lammps/dump',
            type_map=input.type_map,
            input_template=input_template,
        )
    else:
        logger.warn('no available candidates')
        return DummyFuture(GeneralCp2kOutput(dp_data_sets=[], cp2k_outputs=[]))

    # run cp2k tasks
    # group tasks by concurrency
    concurrency = ctx.config.concurrency
    steps_group = [list() for _ in range(concurrency)]
    for i, cp2k_task_dir in enumerate(cp2k_task_dirs):
        steps = steps_group[i % concurrency]
        step = BashStep(
            cwd=cp2k_task_dir,
            cmd=[ctx.config.cp2k_cmd, '-i input.inp 1>> output 2>> output'],
            checkpoint='cp2k',
        )
        steps.append(step)

    jobs = []
    for steps in steps_group:
        if not steps:
            continue
        script = BashScript(
            template=ctx.config.script_template,
            steps=steps,
        )
        job = ctx.executor.submit(script.render(), cwd=tasks_dir)
        jobs.append(job)

    future = GatherJobsFuture(jobs, done_fn=retry_fn(max_tries=2), raise_exception=True)

    # TODO: use a common intermediate data format instead of deepmd/npy
    # ref: https://www.reddit.com/r/comp_chem/comments/10q0isd/whats_the_best_data_format_to_transfer/
    def convert_to_dp_data(_):
        ctx.executor.run_python_fn(cp2k_output_to_dp_data)(
            task_dirs=cp2k_task_dirs,
            output_dir=dp_data_dir,
            type_map=input.type_map)
        dp_data_sets =[
            Artifact(
                executor=ctx.executor.name,
                url=dp_data_dir,
                attrs=dict(fmt='deepmd', set_size=len(cp2k_task_dirs))
            ), # type: ignore
        ]
        cp2k_outputs = [
            Artifact(
                executor=ctx.executor.name,
                url=task_dir,
                attrs=dict(fmt='cp2k')
            )  for task_dir in cp2k_task_dirs  # type: ignore
        ]
        return GeneralCp2kOutput(dp_data_sets=dp_data_sets, cp2k_outputs=cp2k_outputs)
    return MapFuture(future, convert_to_dp_data)


def __make_cp2k_task_dirs():
    """cloudpickle compatible: https://stackoverflow.com/questions/75292769"""
    def make_cp2k_task_dirs(traj_files: List[str],
                            task_dirs: List[str],
                            type_map: List[str],
                            traj_fmt: str,
                            input_template: dict,
                            input_file_name: str = 'input.inp',
                            ):
        import dpdata
        import numpy as np
        from cp2k_input_tools import DEFAULT_CP2K_INPUT_XML
        from cp2k_input_tools.generator import CP2KInputGenerator

        assert len(traj_files) == len(task_dirs), 'len(input_files) != len(task_dirs)'

        cp2k_generator = CP2KInputGenerator(DEFAULT_CP2K_INPUT_XML)

        for i in range(len(traj_files)):
            traj_file = traj_files[i]
            task_dir = task_dirs[i]

            os.makedirs(task_dir, exist_ok=True)
            input_data = copy.deepcopy(input_template)
            system = dpdata.System(traj_file, fmt=traj_fmt, type_map=type_map)

            # format coords
            atom_names = np.array(system['atom_names'])  # type: ignore
            atom_types = system['atom_types']  # type: ignore
            coord_arr = system['coords'][0]  # type: ignore
            kind_arr = atom_names[atom_types]  # type: ignore
            coords = [ str(k) + ' ' + ' '.join(str(x) for x in c)  for k, c in zip(kind_arr, coord_arr) ] # type: ignore

            # format cell
            cell = system['cells'][0]
            cell = np.reshape(cell, [3,3])

            # override input_data
            merge_dict(input_data, {
                'FORCE_EVAL': {
                    'SUBSYS': {
                        'COORD': {
                            '*': coords
                        },
                        'CELL': {
                            'A': list(cell[0, :]),
                            'B': list(cell[1, :]),
                            'C': list(cell[2, :]),
                        }
                    }
                }
            })
            input_text = '\n'.join(cp2k_generator.line_iter(input_data))
            with open(os.path.join(task_dir, input_file_name), 'w') as f:
                f.write(input_text)

    return make_cp2k_task_dirs
make_cp2k_task_dirs = __make_cp2k_task_dirs()


def __cp2k_output_to_dp_data():
    """cloudpickle compatible: https://stackoverflow.com/questions/75292769"""
    def cp2k_output_to_dp_data(task_dirs: List[str], output_dir: str, type_map: List[str]):
        import dpdata
        system = None
        for task_dir in task_dirs:
            _system = dpdata.LabeledSystem(os.path.join(task_dir, 'output'), fmt='cp2k/output', type_map=type_map)
            if system is None:
                system = _system
            else:
                system.append(_system)
        assert system, 'system should not be None, check if task_dirs is empty'
        system.to_deepmd_raw(output_dir)  # type: ignore
        system.to_deepmd_npy(output_dir, set_size = len(system))  # type: ignore

    return cp2k_output_to_dp_data
cp2k_output_to_dp_data = __cp2k_output_to_dp_data()


from ai2_kit.core.executor import Executor
from ai2_kit.core.artifact import Artifact
from ai2_kit.core.log import get_logger
from ai2_kit.core.job import DummyFuture

from typing import List
from io import StringIO
from pydantic import BaseModel
from dataclasses import dataclass
import pandas as pd

import os

logger = get_logger(__name__)

class GeneralThresholdInputConfig(BaseModel):
    f_trust_lo: float
    f_trust_hi: float


@dataclass
class GeneralThresholdContext:
    executor: Executor

@dataclass
class GeneralThresholdOutput:
    candidates: List[Artifact]

@dataclass
class LammpsGeneralThresholdInput:
    config: GeneralThresholdInputConfig
    candidates: List[Artifact]
    model_devi_out_file: str

def lammps_general_threshold(input: LammpsGeneralThresholdInput, ctx: GeneralThresholdContext):

    f_trust_lo = input.config.f_trust_lo
    f_trust_hi = input.config.f_trust_hi
    col_force = 'avg_devi_f'
    logger.info('criteria: %f <= %s < %f ', f_trust_lo, col_force, f_trust_hi)

    for candidate in input.candidates:

        model_devi_out_file = os.path.join(candidate.url, input.model_devi_out_file)
        logger.info('start to analysis file: %s', model_devi_out_file)
        text = ctx.executor.load_text(model_devi_out_file)

        df = pd.read_csv(StringIO(text.lstrip('#')), delim_whitespace=True)
        # layout:
        #        step  max_devi_v  min_devi_v  avg_devi_v  max_devi_f  min_devi_f  avg_devi_f
        # 0        0    0.006793    0.000672    0.003490    0.143317    0.005612    0.026106
        # 1      100    0.006987    0.000550    0.003952    0.128178    0.006042    0.022608

        passed   = df[df[col_force] < f_trust_lo]
        selected = df[(df[col_force] >= f_trust_lo) & (df[col_force] < f_trust_hi)]
        rejected = df[df[col_force] >= f_trust_hi]

        logger.info('result: total: %d, passed: %d, selected: %d, rejected: %d', len(df), len(passed), len(selected), len(rejected))

        candidate.attrs['passed']   = passed.step.tolist()
        candidate.attrs['selected'] = selected.step.tolist()
        candidate.attrs['rejected'] = rejected.step.tolist()

    return DummyFuture(GeneralThresholdOutput(
        candidates=input.candidates,
    ))



from pydantic import BaseModel
from typing import Optional, Dict
from abc import ABC, abstractmethod
import shlex
import os
import re
import shortuuid
import time


from .connector import BaseConnector
from .log import get_logger
from .job import JobFuture, JobState

logger = get_logger(__name__)

class QueueSystemConfig(BaseModel):
    class Slurm(BaseModel):
        sbatch_bin: str = 'sbatch'
        squeue_bin: str = 'squeue'
        scancel_bin: str = 'scancel'
        polling_interval: int = 10

    class Lsf(BaseModel):
        bsub_bin: str = 'bsub'
        bjobs_bin: str = 'bjobs'
        polling_interval: int = 10

    slurm: Optional[Slurm]
    lsf: Optional[Lsf]


class BaseQueueSystem(ABC):

    connector: BaseConnector

    @abstractmethod
    def get_polling_interval(self) -> int:
        ...

    @abstractmethod
    def get_script_suffix(self) -> str:
        ...

    @abstractmethod
    def get_submit_cmd(self) -> str:
        ...

    @abstractmethod
    def get_job_id_pattern(self) -> str:
        ...

    @abstractmethod
    def get_job_state(self, job_id: str, success_indicator_path: str) -> JobState:
        ...

    @abstractmethod
    def cancel(self, job_id: str):
        ...

    def _post_submit(self, job: 'QueueJobFuture'):
        ...

    def submit(self, script: str, cwd: str, name: Optional[str] = None, success_indicator: Optional[str] = None):

        if name is None:
            name = shortuuid.uuid() + self.get_script_suffix()
        quoted_cwd = shlex.quote(cwd)

        if success_indicator is None:
            success_indicator = name + '.success'

        # create script
        script = '\n'.join([
            script,
            '',
            'touch {}'.format(shlex.quote(success_indicator)),
            '',
        ])

        path = os.path.join(cwd, name)
        self.connector.run('mkdir -p {}'.format(quoted_cwd))
        self.connector.dump_text(script, path)

        # submit script
        cmd = "cd {cwd} && {submit_cmd} {script}".format(
            cwd=quoted_cwd,
            submit_cmd=self.get_submit_cmd(),
            script=shlex.quote(name),
        )
        result = self.connector.run(cmd)

        m = re.search(self.get_job_id_pattern(), result.stdout)
        if m is None:
            raise RuntimeError("Unable to parse job id")
        job_id = m.group(1)
        job = QueueJobFuture(self,
                       job_id=job_id,
                       success_indicator=success_indicator,
                       name=name,
                       script=script,
                       cwd=cwd,
                       polling_interval=self.get_polling_interval() // 2,
                       )
        self._post_submit(job)
        return job


class Slrum(BaseQueueSystem):

    config: QueueSystemConfig.Slurm

    _last_states: Optional[Dict[str, JobState]]
    _last_update_at: float = 0

    translate_table = {
        'PD': JobState.PENDING,
        'R': JobState.RUNNING,
        'CA': JobState.CANCELLED,
        'CF': JobState.PENDING,  # (configuring),
        'CG': JobState.RUNNING,  # (completing),
        'CD': JobState.COMPLETED,
        'F': JobState.FAILED,  # (failed),
        'TO': JobState.TIMEOUT,  # (timeout),
        'NF': JobState.FAILED,  # (node failure),
        'RV': JobState.FAILED,  # (revoked) and
        'SE': JobState.FAILED   # (special exit state)
    }

    def get_polling_interval(self):
        return self.config.polling_interval

    def get_script_suffix(self):
        return '.sbatch'

    def get_submit_cmd(self):
        return self.config.sbatch_bin

    def get_job_id_pattern(self):
        # example: Submitted batch job 123
        return r"Submitted batch job\s+(\d+)"

    def get_job_state(self, job_id: str, success_indicator_path: str) -> JobState:
        state = self._get_all_states().get(job_id)
        if state is None:
            cmd = 'test -f {}'.format(shlex.quote(success_indicator_path))
            ret = self.connector.run(cmd, warn=True)
            if ret.return_code:
                return JobState.FAILED
            else:
                return JobState.COMPLETED
        else:
            return state

    def cancel(self, job_id: str):
        cmd = '{} {}'.format(self.config.scancel_bin, job_id)
        self.connector.run(cmd)

    def _post_submit(self, job: 'QueueJobFuture'):
        self._last_states = None

    def _translate_state(self, slurm_state: str) -> JobState:
        return self.translate_table.get(slurm_state, JobState.UNKNOWN)

    def _get_all_states(self) -> Dict[str, JobState]:
        current_ts = time.time()
        if self._last_states is not None and current_ts - self._last_update_at < self.config.polling_interval:
            return self._last_states

        cmd = "{} --noheader --format='%i %t' -u $(whoami)".format(
            self.config.squeue_bin)
        r = self.connector.run(cmd, hide=True)

        states: Dict[str, JobState] = dict()

        for line in r.stdout.split('\n'):
            if not line:  # skip empty line
                continue
            job_id, slurm_state = line.split()

            state = self._translate_state(slurm_state)
            states[job_id] = state
        self._last_update_at = current_ts
        self._last_states = states
        return states


class Lsf(BaseQueueSystem):

    config: QueueSystemConfig.Lsf

    def get_polling_interval(self):
        return self.config.polling_interval

    def get_script_suffix(self):
        return '.lsf'

    def get_submit_cmd(self):
        return self.config.bsub_bin + ' <'

    def get_job_id_pattern(self):
        # example: Job <123> is submitted to queue <small>.
        return r"Job <(\d+)> is submitted to queue"

    # TODO
    def get_job_state(self, job_id: str, success_indicator_path: str) -> JobState:
        return JobState.UNKNOWN

    # TODO
    def cancel(self, job_id: str):
        ...


class QueueJobFuture(JobFuture):

    def __init__(self,
                 queue_system: BaseQueueSystem,
                 job_id: str,
                 success_indicator: str,
                 script: str,
                 cwd: str,
                 name: str,
                 polling_interval=10,
                 ):
        self._queue_system = queue_system
        self._name = name
        self._script = script
        self._cwd = cwd
        self._job_id = job_id
        self._success_indicator = success_indicator
        self._polling_interval = polling_interval
        self._tries = 1
        self._done_state = None

    @property
    def success_indicator_path(self):
        return os.path.join(self._cwd, self._success_indicator)

    def get_job_state(self):
        if self._done_state is not None:
            return self._done_state

        state = self._queue_system.get_job_state(self._job_id, self.success_indicator_path)
        if state.terminal:
            self._done_state = state

        return state

    def redo(self):
        if not self.done():
            raise RuntimeError('Cannot redo an unfinished job!')
        job = self._queue_system.submit(
            script=self._script,
            cwd=self._cwd,
            name=self._name,
            success_indicator=self._success_indicator,
        )
        self._done_state = None
        self._tries += 1
        self._job_id = job._job_id

    def get_tries(self):
        return self._tries

    def is_success(self):
        return self.get_job_state() is JobState.COMPLETED

    def cancel(self):
        self._queue_system.cancel(self._job_id)

    def done(self):
        return self.get_job_state().terminal

    def result(self, timeout: float = float('inf')) -> JobState:

        timeout_ts = time.time() + timeout

        while time.time() < timeout_ts:
            if self.done():
                return self.get_job_state()
            else:
                time.sleep(self._polling_interval)
        else:
            raise RuntimeError(
                'Timeout of polling job: {}'.format(self._job_id))

    def __repr__(self):
        return repr(dict(
            name=self._name,
            cwd=self._cwd,
            job_id=self._job_id,
            success_indicator=self._success_indicator,
            polling_interval=self._polling_interval,
            state=self.get_job_state(),
        ))




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ai2_kit', 'ai2_kit.core', 'ai2_kit.domain', 'ai2_kit.workflow']

package_data = \
{'': ['*']}

install_requires = \
['cloudpickle>=2.2.0,<3.0.0',
 'cp2k-input-tools>=0.8.2,<0.9.0',
 'dpdata>=0.2.13,<0.3.0',
 'fabric>=2.7.1,<3.0.0',
 'fire>=0.4.0,<0.5.0',
 'invoke>=1.7.3,<2.0.0',
 'pandas>=1.5.3,<2.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'ruamel-yaml>=0.17.21,<0.18.0',
 'shortuuid>=1.0.11,<2.0.0']

entry_points = \
{'console_scripts': ['ai2-kit = ai2_kit.main:main']}

setup_kwargs = {
    'name': 'ai2-kit',
    'version': '0.1.0',
    'description': '',
    'long_description': '# ai<sup>2</sup>-kit\n\nA toolkit featured ***a**rtificial **i**ntelligence × **a**b **i**nitio* for computational chemistry research.\n\n*Please be advised that `ai2-kit` is still under heavy development and you should expect things to change often. We encourage people to play and explore with `ai2-kit`, and stay tuned with us for more features to come.*\n\n\n## Feature Highlights\n* A general purpose automated workflow that implements Closed-Loop Learning (CLL) pattern to train Machine Learning Potential (MLP) models.\n* Featured tools for Electrochemistry research:\n    * Automated FEP workflows to train MLP models and calculate redox potential, pKa, solvation, etc.\n* Utilities to execute and manage jobs in local or remote HPC job scheduler.\n* Utilities to simplified automated workflows development with reusable components. \n\n## Installation\n```bash\n# It requires Python >= 3.8\npip install ai2-kit  \n```\n\n## Use Cases\n\n### Train MLP model with CLL workflow\n\n```bash\nai2-kit cll-mlp train-mlp \n```\n\nCCL-MLP workflow implements the Closed-Loop Learning pattern to train MLP models automatically. For each iteration, the workflow will train MLP models and use them to generate new training data for the next round, until the quality of MLP models meets preset criteria. Configurations of each iteration can be updated dynamically to further improve training efficiency.\n\n![cll-mlp-diagram](./doc/img/cll-mlp-diagram.svg)\n\n### Train MLP models for FEP simulation\n\n```bash\nai2-kit ec fep train-mlp\n```\n\n`ec fep` is a dedicated workflow to train MLP models for FEP simulation. Unlike the general purpose `cll-mlp` workflow, `ec fep` workflow uses two different configurations to generate two different labeled structures to train MLP models respectively. And then use the two different models to run FEP simulation.\n\n#### Citation\nIf you use `ec fep` workflow in your research, please cite it:\n> Feng Wang and Jun Cheng, Automated Workflow for Computation of Redox Potentials, Acidity Constants and Solvation Free Energies Accelerated by Machine Learning. J. Chem. Phys, 2022. 157(2), 024103. DOI: https://doi.org/10.1063/5.0098330\n\n\n## TODO\n* Finalize configurations format and provide documents.\n* Provide tools for data format transformation.\n* Provide tools to run MD simulation and properties calculations.\n',
    'author': 'weihong.xu',
    'author_email': 'xuweihong.cn@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from fire import Fire
from ai2_kit.workflow.cll_mlp import cll_train_mlp
from ai2_kit.workflow.fep import fep_train_mlp
from collections import UserDict



# a placeholder to cheating Fire
nil_fn = lambda: None

class Group:

    def __init__(self, items: dict, doc: str = '') -> None:
        self.__doc__ = doc
        self.__dict__.update(items)


toolkit = Group({
    'ec': Group({
        'fep': Group({
            'train-mlp': fep_train_mlp,
        }),
    }),
    'cll-workflow': Group({
        'train-mlp': cll_train_mlp,
    }),
}, doc="Welcome to use ai2-kit!")




def main():
    Fire(toolkit)

if __name__ == '__main__':
    main()



from pydantic import BaseModel
from typing import Optional, Dict
import os


class Artifact(BaseModel):
    executor: Optional[str]
    url: str
    glob: Optional[str]
    attrs: dict = dict()

    referrer: Optional['Artifact']

    def join(self, *paths):
        url = os.path.join(self.url, *paths)
        return Artifact(
            executor=self.executor,
            url=url,
            referrer=self,
        ) # type: ignore

ArtifactMap = Dict[str, Artifact]



from typing import List, Callable, TypeVar, Generic, Optional
from enum import Enum
from abc import abstractmethod
import time

# Copy from parsl
# TODO:

class JobState(bytes, Enum):
    """Defines a set of states that a job can be in"""
    def __new__(cls, value: int, terminal: bool, status_name: str) -> "JobState":
        obj = bytes.__new__(cls, [value])
        obj._value_ = value
        obj.terminal = terminal
        obj.status_name = status_name
        return obj

    value: int
    terminal: bool
    status_name: str

    UNKNOWN = (0, False, "UNKNOWN")
    PENDING = (1, False, "PENDING")
    RUNNING = (2, False, "RUNNING")
    CANCELLED = (3, True, "CANCELLED")
    COMPLETED = (4, True, "COMPLETED")
    FAILED = (5, True, "FAILED")
    TIMEOUT = (6, True, "TIMEOUT")
    HELD = (7, False, "HELD")

class TimeoutError(RuntimeError):
    ...

T = TypeVar('T')

# TODO: refactor design of IFuture for more general use case
# mimic the API of concurrent.future.Future
class IFuture(Generic[T]):

    @abstractmethod
    def done(self) -> bool:
        pass

    @abstractmethod
    def result(self, timeout: Optional[float] = None) -> T:
        pass

class DummyFuture(IFuture[T]):

    def __init__(self, value: T) -> None:
        self.value = value

    def done(self):
        return True

    def result(self, timeout: Optional[float] = None) -> T:
        return self.value


class JobFuture(IFuture[JobState]):

    @abstractmethod
    def get_job_state(self) -> JobState:
        ...

    @abstractmethod
    def cancel(self):
        ...

    @abstractmethod
    def redo(self):
        ...

    @abstractmethod
    def get_tries(self) -> int:
        ...

    @abstractmethod
    def is_success(self) -> bool:
        ...

    def is_retriable(self) -> bool:
        return True


DoneFn = Callable[[JobFuture], bool]

def _default_done_fn(_): return True

def retry_fn(max_tries=2) -> DoneFn:

    def callback(job: JobFuture):
        if job.is_success():
            return True
        if job.get_tries() >= max_tries:
            return True
        if job.is_retriable():
            job.redo()
            return False

    return callback  # type: ignore


class GatherJobsFuture(IFuture[List[JobState]]):

    def __init__(self, jobs: List[JobFuture],
                 done_fn: DoneFn = _default_done_fn,
                 raise_exception = True,
                 polling_interval = 10,
                 ):
        self._jobs = jobs
        self._done_fn = done_fn
        self._raise_exception = raise_exception
        self._polling_interval = polling_interval


    def done(self):
        all_done = True
        for job in self._jobs:
            all_done = all_done and job.done() and self._done_fn(job)
        return all_done

    def result(self, timeout: Optional[float] = None):
        if timeout is None:
            timeout = float('inf')
        timeout_ts = time.time() + timeout

        while time.time() < timeout_ts:
            if self.done():
                if self._raise_exception:
                    failed_jobs = [
                        job for job in self._jobs if not job.is_success()]
                    if failed_jobs:
                        raise RuntimeError(
                            'Soem jobs are failed: {}'.format(failed_jobs))
                return [job.result() for job in self._jobs]

            else:
                time.sleep(self._polling_interval)
        else:
            raise TimeoutError('Wait for jobs timeout!')


NT = TypeVar('NT')

MapFn = Callable[[T], NT]

class MapFuture(IFuture[NT]):

    def __init__(self, future: IFuture[T], map_fn: MapFn):
        self._future = future
        self._map_fn = map_fn

    def done(self):
        return self._future.done()

    def result(self, timeout=None) -> NT:
        result = self._future.result(timeout)
        return self._map_fn(result)

def map_future(future: IFuture, value: T):
    return MapFuture[T](future, lambda _ : value)
